#include "config.h"
#include "system.h"
#include "coretypes-wrapper.h"
#include "input.h"
#include "tree.h"
#include "errors.h"

#include "tokens.h"
#include "lex.h"
#include "asdf-tree.h"

#include "hashtab.h"
#include "asdf-hash.h"
#include "nl-eval.h"
#include "ggc.h"

#define NEXT_THRESHOLD(table) (((table)->impl->length * 192)/256)
#define IMPL_SIZE(length) ((sizeof(struct asdf_hashtab_impl)) - sizeof(tree)*1 + sizeof(tree)*(length) )

static void dump_statistics( struct asdf_hashtab *table );
static unsigned int next_prime( unsigned int num );

/*
Table of prime numbers 2^n+a, 2<=n<=30.

from st.c
"st.c 5.1 89/12/14 Crucible"
*/
static unsigned int primes[] = {
	8 + 3,
	16 + 3,
	32 + 5,
	64 + 3,
	128 + 3,
	256 + 27,
	512 + 9,
	1024 + 9,
	2048 + 5,
	4096 + 3,
	8192 + 27,
	16384 + 43,
	32768 + 3,
	65536 + 45,
	131072 + 29,
	262144 + 3,
	524288 + 21,
	1048576 + 7,
	2097152 + 17,
	4194304 + 15,
	8388608 + 9,
	16777216 + 43,
	33554432 + 35,
	67108864 + 15,
	134217728 + 29,
	268435456 + 3,
	536870912 + 11,
	1073741824 + 85,
	0
};


static 
unsigned int next_prime( unsigned int num )
{
  int max = sizeof(primes)/sizeof(int);

  int i;

  for ( i=0; i<max; i++ ) {
    if ( primes[i] > num )
      return primes[i];
  }

  internal_error("too large prime\n");
}

static void 
dump_statistics( struct asdf_hashtab *table ) 
{
  int i;

  for ( i=0; i<table->impl->length; i++ ) {
    tree t;
    int count = 0;

    for ( t = table->impl->entries[i]; t; t=TREE_CHAIN(t) ) {
      count++;
    }
    fprintf( stderr, "%d: %d\n", i, count );
  }

  fputs("\n\n", stderr);
}

void 
asdf_hashtab_foreach( struct asdf_hashtab *table, 
			   asdf_hashtab_cb cb, 
			   asdf_htab_cb_t arg)
{
  int i;

  for ( i=0; i<table->impl->length; i++ ) {
    tree t;

    for ( t = table->impl->entries[i]; t; t=TREE_CHAIN(t) ) {
      (*cb)(t,arg);
    }
  }
}

struct asdf_hashtab *
asdf_hashtab_create(int init_size)
{
  struct asdf_hashtab *table = ggc_alloc( sizeof(struct asdf_hashtab) );

  int size = next_prime(init_size);
  int i;

  table->impl = ggc_alloc( IMPL_SIZE(size) );
  table->impl->length = size;

  table->num_ent = 0;
  table->threshold = NEXT_THRESHOLD(table) ;
  
  for ( i=0; i<size; i++ ) {
    table->impl->entries[i] = NULL;
  }

  return table;
}

unsigned int 
nl_hash( tree key )
{
  switch ( TREE_CODE(key) ) {
  case IDENTIFIER_NODE: {
    const char *id = IDENTIFIER_POINTER(key);

    return htab_hash_string(id) + 1;
  }

  case NL_STRING: {
    const char *str = NL_STRING_POINTER(key);
    return htab_hash_string(str);
  }

  case INTEGER_CST: 
    return (unsigned int)TREE_INT_CST_LOW(key);

  default:
    return (unsigned int)key;
  }

  return 0;
}

int
nl_eql( tree t1, tree t2 ) {

  if ( TREE_CODE(t1) != TREE_CODE(t2) ) 
    return 0;

  switch ( TREE_CODE(t1) ) {
  case INTEGER_CST:
    return (int)(TREE_INT_CST_LOW(t1)) == (int)(TREE_INT_CST_LOW(t2));

  case NL_STRING: {
    const char *s1 = NL_STRING_POINTER(t1),
      *s2 = NL_STRING_POINTER(t2);

    return !(strcmp(s1,s2));
  }

  case IDENTIFIER_NODE: {
    const char *s1 = IDENTIFIER_POINTER(t1),
      *s2 = IDENTIFIER_POINTER(t2);
    
    return !(strcmp(s1,s2));
  } 


  default:
    return t1 == t2;

  }

}

void
asdf_hashtab_rehash( struct asdf_hashtab *tab, int request_size )
{
  int size = next_prime( request_size );
  int i;

  struct asdf_hashtab_impl *new_impl = ggc_calloc( IMPL_SIZE(size), 1 );
  
  for ( i=0; i<tab->impl->length; i++ ) {
    tree t = tab->impl->entries[i];

    while ( t ) {
      tree next = TREE_CHAIN(t);
      int hash = nl_hash(NL_ENTRY_KEY(t))%size;

      TREE_CHAIN(t) = new_impl->entries[hash];
      new_impl->entries[hash] = t;

      t = next;
    }
    
  }

  tab->impl = new_impl;
  tab->impl->length = size;
  tab->threshold = NEXT_THRESHOLD(tab);
}

static void
add_newnode( struct asdf_hashtab *table, tree node, int hash )
{
  TREE_CHAIN(node) = table->impl->entries[hash];
  table->impl->entries[hash] = node;

  table->num_ent++;

  if ( table->num_ent > table->threshold ) {
    asdf_hashtab_rehash( table, table->num_ent * 3 );
  }
}

tree
asdf_hashtab_lookup( struct asdf_hashtab *table, tree key, int add )
{
  unsigned int hash = nl_hash(key)%table->impl->length;
  tree node;

  for ( node = table->impl->entries[hash]; node!=NULL; node=TREE_CHAIN(node) ) {
    if ( nl_eql(NL_ENTRY_KEY(node),key) == 1 )
      return node;
  }

  if ( add ) {
    tree new_node = make_nl_entry(key,NULL);
    add_newnode( table, new_node, hash );
    return new_node;
  }

  return NULL;
}

tree
asdf_hashtab_insert( struct asdf_hashtab *table, tree key, tree value )
{
  int hash = nl_hash(key)%table->impl->length;
  tree node = make_nl_entry( key, value );

  add_newnode( table, node, hash );

  return node;
}

