#include "config.h"
#include "system.h"
#include "coretypes-wrapper.h"
#include "input.h"
#include "tree.h"
#include "errors.h"

#include "tokens.h"
#include "lex.h"
#include "asdf-tree.h"

#include "nl-eval.h"

#include "hashtable.h"
#include "ggc.h"

tree global_namespace;
tree current_namespace;
struct nl_environment *current_env;
struct nl_ggc_prot *nl_prot_root = NULL;

static tree eval_def_stmt( tree );
static void eval_var_stmt( tree );
static tree eval_if_stmt( tree );
static tree eval_while_stmt( tree );
static tree eval_for_stmt( tree );
static tree eval_id( tree );
static tree eval_binary_op( tree );
static tree eval_compare_expr( tree );
static tree eval_modify_expr( tree );
static tree eval_stmts( tree );
static tree eval_context_block( tree );
static tree eval_block( tree );
static tree eval_ns_block( tree );
static tree eval_scope_expr( tree );
static tree eval_call_expr( tree );

static tree call_function( tree func, tree args );
static void redefine_const_entry(tree entry, const char *id);

tree
nl_eval (expr)
     tree expr;
{
  switch ( TREE_CODE(expr) ) {

  case NL_SPECIAL_VALUE: 
    return expr;

  case IDENTIFIER_NODE:
    return eval_id( expr );

  case PLUS_EXPR:
  case MINUS_EXPR:
  case MULT_EXPR:
    return eval_binary_op( expr );

  case GT_EXPR:
  case GE_EXPR:
  case LT_EXPR:
  case LE_EXPR:
  case NE_EXPR:
  case EQ_EXPR:
    return eval_compare_expr( expr );

  case INTEGER_CST:
  case NL_STRING:
    return expr;

  case MODIFY_EXPR:
    return eval_modify_expr( expr );

  case NL_DEF_STMT:
    return eval_def_stmt( expr );
      
  case NL_VAR_STMT:
    eval_var_stmt( expr );
    return NULL;

  case IF_STMT:
    return eval_if_stmt( expr );

  case WHILE_STMT:
    return eval_while_stmt( expr );

  case FOR_STMT:
    return eval_for_stmt( expr );
    
  case NL_EXPR_STMT:
    input_line = NL_EXPR_STMT_SOURCE_LINE( expr );
    input_filename = NL_EXPR_STMT_SOURCE_FILE( expr );
    return nl_eval( NL_EXPR_STMT_EXPR(expr) );

  case ASDF_BLOCK:
    return eval_context_block( expr );

  case NL_NAMESPACE_BLOCK:
    return eval_ns_block( expr );

  case NL_SCOPE_EXPR:
    return eval_scope_expr( expr );

  case CALL_EXPR:
    return eval_call_expr( expr );

  default:
    internal_error( "nl_eval. appears '%s'.", tree_code_name[TREE_CODE(expr)] );
    break;
  }

  return NULL;
}

static tree
call_function( tree func, tree args )
{
  tree ns = nl_namespace_create();
  tree fargs = NL_FUNCTION_ARGS(func);

  for ( ; args!=nl_nil; args=NL_CONS_TL(args) ) {
    tree argid;

    if ( fargs == nl_nil ) {
      error("Wrong number of argment");
      exit(1);
    }

    argid = NL_CONS_HD(fargs);

    nl_namespace_define( ns, argid, NL_CONS_HD(args) );
    
    fargs = NL_CONS_TL(fargs);
  }

  if ( fargs != nl_nil ) {
    error("Wrong number of argments");
    exit(1);
  }

  set_environment( NL_FUNCTION_DEFINED_ENV(func) );
  push_namespace( ns );

  eval_block( NL_FUNCTION_BODY(func) );

  pop_environment();

  return nl_retval;
}

static tree
eval_call_expr( tree expr )
{
  tree func = TREE_OPERAND( expr, 0 );
  tree args = TREE_OPERAND( expr, 1 );

  nl_native_func_t nfunc;
  int argnum;

  tree e = NULL;
  tree a;
  int len=0;

  func = nl_eval( func );

  if ( TREE_CODE(func) == NL_NATIVE_FUNCTION ) {
    /* nothing */
  } else if ( TREE_CODE(func) == NL_FUNCTION ) {
    return call_function( func, args );
  } else {
    error("invalid function value.");
    exit( 1 );
  }    

  for ( a=args; a!=nl_nil; a=TREE_CHAIN(a) ) {
    tree cons = make_node(NL_CONS);
    
    NL_CONS_HD(cons) = nl_eval( NL_CONS_HD(a) );
    NL_CONS_TL(cons) = e;
    
    e = cons;

    len++;
  }

  nfunc = NL_NATIVE_FUNCTION_FUNC(func);
  argnum = NL_NATIVE_FUNCTION_ARGNUM(func);
  
  if ( (argnum>=0) && (len!=argnum) ) {
    error("Wrong number of argments.");
    exit( 1 );
  }
  return (*nfunc)( e );
}

/* a::b 
 *
 * if expr reference as left value like:
 *
 *** a::b = 40;
 *
 * use eval_scope_expr_as_lval;
 */
static tree
eval_scope_expr( tree expr )
{
  tree lval = TREE_OPERAND( expr, 0 );
  tree rval = TREE_OPERAND( expr, 1 );
  tree ret;

  tree ns = nl_eval( lval );

  if ( TREE_CODE(ns) != NL_NAMESPACE ) {
    error("reference to not namespace value in scope_expr.");
    exit( 1 );
  }

  push_environment( ns );

  ret = nl_eval( rval );

  pop_environment();

  return ret;
}

static tree
eval_ns_block( tree block )
{
  tree id = NL_NAMESPACE_BLOCK_NAME(block);
  tree ns = nl_lookup( id );
  tree ret;

  if ( ns == NULL ) {
    ns = nl_namespace_create();
    nl_namespace_define( current_namespace, id, ns );
    TREE_CONSTANT( ns ) = 1;
  } else if ( TREE_CODE(ns) != NL_NAMESPACE ) {
    error("'%s' is not namespace.");
    exit( 1 );
  }

  push_namespace( ns );

  input_line = NL_NAMESPACE_BLOCK_SOURCE_LINE(block);
  input_filename = NL_NAMESPACE_BLOCK_SOURCE_FILE(block);

  ret = copy_node( block );
  NL_NAMESPACE_BLOCK_BODY(ret) = eval_block( NL_NAMESPACE_BLOCK_BODY(block) );
  pop_namespace();

  return ret;
}

static tree
eval_context_block( tree block )
{
  tree ns = nl_namespace_create();
  tree ret;

  push_namespace( ns );
  ret = eval_block( block );
  pop_namespace( );
  
  return ret;
}

/* todo: setup context */
static tree
eval_block( tree block )
{
  tree stmts = ASDF_BLOCK_BODY( block );
  tree body = eval_stmts( stmts );

  tree ret = copy_node( block );
  ASDF_BLOCK_BODY(ret) = body;

  return ret;
}

static tree
eval_scope_expr_as_lval( tree expr )
{
  tree lval = TREE_OPERAND(expr,0);
  tree rval = TREE_OPERAND(expr,1);

  tree id;
  tree ret;
  
  tree ns = nl_eval( lval );

  if ( TREE_CODE(ns) != NL_NAMESPACE ) {
    error("references not namespace value as namespace.");
    exit(1);
  }

  push_environment( ns );
  
  if ( TREE_CODE(rval) == IDENTIFIER_NODE ) {
    id = rval;
  } else {
    id = nl_eval(rval);

    if ( TREE_CODE(rval) != IDENTIFIER_NODE ) {
      error("references not identifier value as left value.");
      exit(1);
    }
  }

  ret = nl_lookup_entry( id );

  if ( ret == NULL ) {
    error("identifier '%s' is not defined.", IDENTIFIER_POINTER(id) );
    exit( 1 );
  }
  
  pop_environment();
  
  return ret;
}

static tree
eval_modify_expr( tree e )
{
  tree lval = TREE_OPERAND( e, 0 );
  tree rval = nl_eval(TREE_OPERAND( e, 1 ));
  tree id = NULL;
  tree entry;

  if ( TREE_CODE(lval) == NL_SCOPE_EXPR ) {

    /* if error is occured, this function does not return */
    entry = eval_scope_expr_as_lval(lval);

  } else if ( TREE_CODE(lval) == IDENTIFIER_NODE ) {

    id = lval;
    entry = nl_lookup_entry( lval );

  } else {

    error("assign left value is not identifier");
    exit( 1 );

  }

  if ( entry == NULL ) {
    error("identifier '%s' is not defined.", IDENTIFIER_POINTER(id) );
    exit( 1 );
  }

  if ( TREE_CONSTANT(entry) == 1 ) {
    redefine_const_entry( entry, IDENTIFIER_POINTER(id) );
  }

  NL_ENTRY_VALUE(entry) = rval;
  return rval;
}

static tree
eval_compare_expr( tree e )
{
  tree l = nl_eval( TREE_OPERAND(e,0) );
  tree r = nl_eval( TREE_OPERAND(e,1) );
  enum tree_code op = TREE_CODE(e);

  int val;

  if ( TREE_CODE(l) != TREE_CODE(r) ) {
    error("compare with expression has different type.");
    exit( 1 );
  }

  if ( TREE_CODE(l) == INTEGER_CST ) {
    int lval = TREE_INT_CST_LOW(l);
    int rval = TREE_INT_CST_LOW(r);

    switch (op) { 
    case EQ_EXPR:
      val = (lval==rval);
      break;

    case NE_EXPR:
      val = (lval!=rval);
      break;

    case GT_EXPR:
      val = (lval>rval);
      break;

    case GE_EXPR:
      val = (lval>=rval);
      break;

    case LT_EXPR:
      val = (lval<rval);
      break;

    case LE_EXPR:
      val = (lval<=rval);
      break;

    default:
      internal_error("compare by nanika");
    }

  } else if ( (op==EQ_EXPR) ) {
    
    val = nl_eql(l,r);

  } else if ( (op==NE_EXPR) ) {
    
    val = ! nl_eql(l,r);

  } else {
    error("comparation between '%s' and '%s'.", 
	  tree_code_name[TREE_CODE(l)], tree_code_name[TREE_CODE(r)]);
    exit(1);
  }

  if ( val ) {
    return nl_true;
  } else {
    return nl_false;
  }
}

static tree
eval_binary_op( tree e )
{
  /* fixme: protect */

  tree l = nl_eval( TREE_OPERAND(e,0) );
  tree r = nl_eval( TREE_OPERAND(e,1) );

  int lval, rval;
  int retval;

  if ( TREE_CODE(l) != INTEGER_CST ||
       TREE_CODE(r) != INTEGER_CST ) {
    error("binary_op apply to not integer value");
    exit( 1 );
  }

  lval = TREE_INT_CST_LOW(l);
  rval = TREE_INT_CST_LOW(r);

  switch ( TREE_CODE(e) ) {
  case PLUS_EXPR:
    retval = lval+rval;
    break;
  case MINUS_EXPR:
    retval = lval-rval;
    break;
  case MULT_EXPR:
    retval = lval*rval;
    break;

  default:
    error("eval_binaray_op internal error");
    abort ();
  }

  return build_int_2( retval, 0 );
}

static tree
eval_id( tree id )
{
  const char *name = IDENTIFIER_POINTER( id );
  tree value = nl_lookup( id );

  if ( value == NULL ) {
    warning( "undeifned variable '%s'.", name );
    return nl_nil;
  }

  return value;  
}


static tree
eval_def_stmt (stmt)
     tree stmt;
{
  tree func;
  tree id = NL_DEF_STMT_ID(stmt);
  tree args = NL_DEF_STMT_ARGS(stmt);
  tree body = NL_DEF_STMT_BODY(stmt);

  tree entry;

  func = build_function( args, body );

  entry = nl_namespace_define( current_namespace, id, func );

  if ( TREE_CONSTANT(entry) ) {
    redefine_const_entry( entry, IDENTIFIER_POINTER(id) );
  }

  NL_ENTRY_DEFINE_FILE(entry) = input_filename;
  NL_ENTRY_DEFINE_LINE(entry) = input_line;

  TREE_CONSTANT(entry) = 1;

  return nl_nil;
}

static tree
eval_if_stmt( tree stmt )
{
  /* fixme: location */

  tree cond = nl_eval( IF_COND(stmt) );
  tree ret;

  if ( ! NL_IS_FALSE(cond) ) {
    ret = nl_eval( THEN_CLAUSE(stmt) );
  } else if ( ELSE_CLAUSE(stmt) ) {
    ret = nl_eval( ELSE_CLAUSE(stmt) );
  } else {
    ret = nl_nil;
  }
  
  return ret;
}
 
static tree
eval_while_stmt( tree stmt )
{
  tree cond_expr = WHILE_COND(stmt);
  tree body = WHILE_BODY(stmt);

  while ( 1 ) {
    tree cond = nl_eval( cond_expr );
    
    if ( NL_IS_FALSE(cond) )
      break;

    nl_eval( body );
  }

  return nl_nil;
}    

static tree
eval_for_stmt( tree stmt )
{
  tree cond_expr = FOR_COND(stmt);
  tree iterate_expr = FOR_EXPR(stmt);
  tree body = FOR_BODY(stmt);

  nl_eval( FOR_INIT_STMT(stmt) );

  while ( 1 ) {
    tree cond = nl_eval( cond_expr );

    if ( NL_IS_FALSE(cond) ) 
      break;

    nl_eval( iterate_expr );
    nl_eval( body );
  }

  return nl_nil;
}

static void
eval_var_stmt (stmt)
     tree stmt;
{
  tree id = NL_VAR_STMT_ID( stmt );
  const char *name = IDENTIFIER_POINTER( NL_VAR_STMT_ID(stmt) );
  tree init = NL_VAR_STMT_INIT(stmt);
  tree entry;

  input_filename = NL_VAR_STMT_SOURCE_FILE(stmt);
  input_line = NL_VAR_STMT_SOURCE_LINE(stmt);

  if ( init ) {
    tree val = nl_eval(init);

    entry = nl_namespace_define( current_namespace, id, val );
  } else {
    entry = nl_namespace_define( current_namespace, id, nl_nil );
  }

  if ( TREE_CONSTANT(entry) ) {
    redefine_const_entry( entry, name );
  }

  NL_ENTRY_DEFINE_FILE(entry) = input_filename;
  NL_ENTRY_DEFINE_LINE(entry) = input_line;

  TREE_CONSTANT(entry) = TREE_CONSTANT(stmt);
}

static tree
eval_stmts (stmts)
     tree stmts;
{
  tree s;
  tree ret = NULL;
  tree *rp = &ret;
  NL_BEGIN_PROTECT;

  NL_PROTECT( ret );

  for ( s=stmts; s!=NULL; s=NL_CONS_TL(s) ) {
    tree val = nl_eval( NL_CONS_HD(s) );

    if ( (val == NULL) || (val == nl_nil) ) 
      continue;

    if ( EXPR_P(val) || 
	 TREE_CODE_CLASS(TREE_CODE(val)) == 'c' ) {
      tree e = val;
      
      val = make_node( NL_EXPR_STMT );
      NL_EXPR_STMT_EXPR(val) = e;
      NL_EXPR_STMT_LOCUS(val) = ggc_alloc(sizeof(location_t));
      NL_EXPR_STMT_SOURCE_LINE(val) = input_line;
      NL_EXPR_STMT_SOURCE_FILE(val) = input_filename;
    }

    *rp = make_node(NL_CONS);
    NL_CONS_HD(*rp) = val;

    rp = &(NL_CONS_TL(*rp));
  }

  *rp = NULL;

  NL_END_PROTECT;

  return ret;
}

tree
nl_eval_program (tree program)
{
  return eval_stmts( program );
}

void 
nl_init(void)
{
  current_namespace = global_namespace = nl_namespace_create();

  current_env = ggc_alloc( sizeof(struct nl_environment) );
  current_env->namespace_list = build_tree_list( NULL, current_namespace );
  
  nl_true = make_node( NL_SPECIAL_VALUE );
  nl_false = make_node( NL_SPECIAL_VALUE );
  nl_nil = make_node( NL_SPECIAL_VALUE );

  nl_init_std();
}


tree nl_true;
tree nl_false;
tree nl_nil;
tree nl_retval;

static void
redefine_const_entry( tree entry, const char *id )
{
  error("redefine constant variable '%s'.", id);

  input_line = NL_ENTRY_DEFINE_LINE(entry);
  input_filename = NL_ENTRY_DEFINE_FILE(entry);
  error("'%s' was defined here.", NL_ENTRY_KEY(entry) );

  exit(1);
}

tree
build_function( tree args, tree body )
{
  tree func = make_node( NL_FUNCTION );

  NL_FUNCTION_ARGS(func) = args;
  NL_FUNCTION_BODY(func) = body;
  NL_FUNCTION_DEFINED_ENV(func) = current_env->namespace_list;

  NL_FUNCTION_LOCUS(func) = ggc_alloc( sizeof(location_t) );

  NL_FUNCTION_LINE(func) = input_line;
  NL_FUNCTION_FILE(func) = input_filename;

  return func;
}
